import sys

from scrub import scrubme
from scrub.utils import diff_results
from scrub.utils import scrub_utilities


help_message = ('User Documentation: https://nasa.github.io/scrub\n\n' +
                'run\n' +
                scrubme.main.__doc__ + '\n\n'
                'diff\n' +
                diff_results.diff.__doc__ + '\n\n'
                'get-conf\n' +
                scrub_utilities.create_conf_file.__doc__ + '\n')


def main():
    """Console script for SCRUB."""

    if len(sys.argv) <= 1:
        print(help_message)
    else:
        if 'run' in sys.argv or 'run-all' in sys.argv:
            # Run analysis
            scrubme.parse_arguments()

        elif 'run-tool' in sys.argv:
            # Get the tool name
            tool = sys.argv[sys.argv.index('--module') + 1].split('_')[-1]
            insert_index = sys.argv.index('--module')

            # Update the arguments
            sys.argv[1] = 'run'
            sys.argv[insert_index] = '--tool'
            sys.argv[insert_index + 1] = tool

            # Run analysis
            scrubme.parse_arguments()

        elif 'diff' in sys.argv:
            # Run analysis
            diff_results.parse_arguments()

        elif 'get-conf' in sys.argv:
            # Run analysis
            scrub_utilities.create_conf_file()

        else:
            print(help_message)

    return 0


if __name__ == "__main__":
    sys.exit(main())



import os
import shutil


def clean_subdirs(root_dir):
    """This function cleans all sub-directories under the specified root directory.

    Inputs:
        - root_dir: Absolute path to the root directory of interest [string]
    """

    # Find all the files and directories of interest
    root_sub_dir = True
    for root, dir_names, file_names in os.walk(root_dir):
        # Update the dir list only the first time
        if root_sub_dir:
            # Remove .scrub from the list, if it exists
            if '.scrub' in dir_names:
                dir_names.remove('.scrub')

            # Update the flag
            root_sub_dir = False

        # Remove any .scrub directories
        for dir_name in dir_names:
            if '.scrub' in dir_name:
                dir_path = os.path.join(root, dir_name)
                shutil.rmtree(dir_path)


def clean_directory(directory):
    """This function removes previous SCRUB data products.

    Inputs:
        - directory: Full path to the top-level source code directory [string]
    """

    # Remove the root directory
    if os.path.exists(directory + '/.scrub'):
        shutil.rmtree(directory + '/.scrub')

    # Remove all of the sub-directories
    clean_subdirs(directory)



import os
import sys
import glob
import datetime
import shutil
from scrub.tools.parsers import translate_results


def parse_warnings(input_dir):
    # Initialize variables
    source_dir = os.path.dirname(input_dir)
    output_dir = os.path.normpath(input_dir + '/csv_output')

    # Make the output directory if it doesn't already exist
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.mkdir(output_dir)

    # Get a list of all the scrub files in the directory
    input_files = glob.glob(input_dir + '/*.scrub')

    # Create the output file path
    output_file = (output_dir + '/' + datetime.datetime.utcnow().strftime("%m-%d-%Y") + '.csv')

    # Write the results out to the csv file
    with open(output_file, 'w') as output_fh:
        # Write out the header data
        output_fh.write('ID,Query,Description,Priority,File,Line\n')

        # Iterate through every file
        for input_file in input_files:

            # Parse the SCRUB files
            scrub_results = translate_results.parse_scrub(input_file, source_dir)

            # Iterate through every result
            for result in scrub_results:
                output_fh.write('{},{},{},{},{},{}\n' .format(result.get('id'), result.get('query'),
                                                              ' '.join(result.get('description')).replace(',', ''),
                                                              result.get('priority'),
                                                              os.path.relpath(result.get('file'), source_dir),
                                                              result.get('line')))


if __name__ == '__main__':
    parse_warnings(sys.argv[1])



import re
import os
import logging
import pwd
import traceback
import shutil
import subprocess
from scrub.utils import scrub_utilities
from scrub.utils.filtering import create_file_list


def login(bin_dir, login_flags):
    """This function logs in users to Collaborator.

    Inputs:
        - bin_dir: Absolute path to the Collaborator ccollab client [string]
        - login_flags: Command flags to be passed to the Collaborator command 'ccollab login' [string]
    """

    # Construct the call string
    if bin_dir == '':
        call_string = 'ccollab --quiet login ' + login_flags
    else:
        call_string = bin_dir + '/ccollab --quiet login ' + login_flags

    # Execute the command
    scrub_utilities.execute_command(call_string, os.environ.copy(), None, True)


def check_login(bin_dir):
    """This function checks to see if a user is logged in to the Collaborator server.

    Inputs:
        - bin_dir: Absolute path to the Collaborator ccollab client [string]
    """

    # Construct the call string
    if bin_dir == '':
        call_string = 'ccollab info'
    else:
        call_string = bin_dir + '/ccollab info'

    # Execute the command
    proc = subprocess.Popen(call_string, shell=True, env=os.environ.copy(), stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, encoding='utf-8')
    std_out, std_err = proc.communicate()

    # Check to see if a user is logged in
    if std_out.find('Connected as:') >= 0:
        authenticated = True
    else:
        authenticated = False

    return authenticated


def execute_ccollab(bin_dir, subcommand):
    """This function executes a given ccollab command.

    Inputs:
        - bin_dir: Absolute path to the Collaborator ccollab client [string]
        - subcommand: Command string and custom flags to be executed [string]
    """

    # Construct the call string
    if bin_dir == '':
        call_string = 'ccollab --no-browser --scm none ' + subcommand
    else:
        call_string = bin_dir + '/ccollab --no-browser --scm none ' + subcommand

    # Execute the command
    scrub_utilities.execute_command(call_string, os.environ.copy())


def get_review_id(log_file):
    """This function finds the review ID from the Collaborator log file.

    Inputs:
        - log_file: Absolute path to the Collaborator log file [string]

    Outputs:
        review_id: Review ID number found in the Collaborator log file [string]
    """

    # Initialize variables
    review_id = 0

    # Import the log data
    with open(log_file, 'r') as input_fh:
        review_id_data = input_fh.readlines()

    # Find the review ID
    for line in review_id_data:
        if 'Successfully created review' in line:
            review_id = re.split(' ', line.strip())[-1].replace('.', '')

    return review_id


def get_defects(scrub_file):
    """This function parses the input file and creates an internal dictionary of findings.

    Inputs:
        - scrub_file: Absolute path to the SCRUB-formatted results file [string]

    Outputs:
        defect_list: List of defects found within the results file [list of dict]
    """

    # Initialize variables
    special_chars_dict = {'"': "&quot;",
                          "'": "&apos;",
                          "<": "&lt;",
                          ">": "&gt;",
                          "&": "&amp;"}
    defect_list = []

    # Import the defects from the file of interest
    with open(scrub_file, 'r') as fh:
        results = fh.read()

    # Split the results into defects
    defects = results.split('\n\n')

    # Extract defect info
    for raw_defect in defects:
        if raw_defect:
            # Initialize the dict
            defect = {'id': None,
                      'file': None,
                      'line': None,
                      'priority': None,
                      'description': None,
                      'tool': None}

            # Replace any whitespace with single space
            raw_defect = re.sub(r'\s+', ' ', raw_defect)

            # Get tool abbrv, defect num, severity, file name, line num, description
            raw_defect = re.match(r'(\D+\d*\D+)(\d*) <(\D+)> :(\S+):(\d+): (.+)', raw_defect)

            if 'p10' in os.path.basename(scrub_file):
                defect.update({'tool': 'p10_' + raw_defect.group(1)})
            else:
                defect.update({'tool': raw_defect.group(1)})
            defect_num = raw_defect.group(2)

            defect.update({'id': defect.get('tool') + defect_num})
            defect.update({'priority': raw_defect.group(3)})
            defect.update({'file': raw_defect.group(4)})
            defect.update({'line': raw_defect.group(5)})
            defect.update({'description': raw_defect.group(6)})

            # Convert special chars in description
            for key in special_chars_dict.keys():
                defect.update({'description': defect.get('description').replace(key, special_chars_dict[key])})

            # Add the defect to the list
            defect_list.append(defect)

    return defect_list


def construct_xml_global_options(url, username):
    """This function constructs the global options portion of the XML file.

    Inputs:
        - url: URL of the Collaborator server [string]
        - username: Username for the Collaborator user [string]

    Outputs:
        - global_options_string: A string containing the global options [string]
    """

    # Create the global options string
    global_options_string = ('    <global-options>\n' +
                             '        <no-browser/>\n' +
                             '        <scm>none</scm>\n' +
                             '        <url>{}</url>\n'.format(url) +
                             '        <user>{}</user>\n'.format(username) +
                             '    </global-options>\n')

    return global_options_string


def create_batch_xml_file_upload(output_file, file_list, review_id, url, username, source_root):
    """This function creates the XML file that is used to upload review files to Collaborator.

    Inputs:
        - output_file: Absolute path to the desired XML output file location [string]
        - file_list: List of files to be uploaded to Collaborator [list of strings]
        - review_id: ID of review to upload files [string]
        - url: URL of the Collaborator server [string]
        - username: Username for the Collaborator user [string]
        - source_root: Absolute path to the source code root directory [string]
    """

    with open(output_file, 'w') as output_fh:
        # Start the XML
        output_fh.write('<batch-commands>\n')

        # Write out the global options
        output_fh.write('{}'.format(construct_xml_global_options(url, username)))

        # Write out all of the file upload commands
        output_fh.write('    <addfiles>\n')
        output_fh.write('        <review>{}</review>\n'.format(review_id))
        output_fh.write('        <relative-to>{}</relative-to>\n'.format(source_root))
        output_fh.write('        <upload-comment>Initial SCRUB upload</upload-comment>\n')

        for file in file_list:
            output_fh.write('        <file-path>{}</file-path>\n'.format(file.strip()))

        output_fh.write('    </addfiles>\n')

        # Close out the file
        output_fh.write('</batch-commands>\n')


def create_batch_xml_defect_upload(output_file, file_list, defect_list, review_id, finding_level, url, username):
    """This function creates the XML file that is used to upload review defects to Collaborator.

    Inputs:
        - output_file: Absolute path to the desired XML output file location [string]
        - defect_list: List of defects to be uploaded to the Collaborator review [list of dicts]
        - review_id: ID of review to upload files [string]
        - finding_level: Should findings be uploaded as comments or defects? ['comment'/'defect']
        - url: URL of the Collaborator server [string]
        - username: Username for the Collaborator user [string]
    """

    with open(output_file, 'w') as output_fh:
        # Start the XML
        output_fh.write('<batch-commands>\n')

        # Write out the global options
        output_fh.write('{}'.format(construct_xml_global_options(url, username)))
        
        # Track if we're including any warnings
        uploaded_warnings = False

        # Write out every defect for files of interest
        for defect in defect_list:
            if defect.get('file') in file_list:
                uploaded_warnings = True
                if finding_level.lower() == 'defect':
                    # Parse the priority level
                    if defect.get('priority') == 'High':
                        severity = 'Major'
                    elif defect.get('priority') == 'Med':
                        severity = 'Moderate'
                    else:
                        severity = 'Minor'

                    output_fh.write('    <admin_review_defect_create>\n')
                    output_fh.write('        <custom-field>Severity={}</custom-field>\n'.format(severity))
                else:
                    output_fh.write('    <admin_review_comment_create>\n')

                output_fh.write('        <review>{}</review>\n'.format(review_id))
                output_fh.write('        <file>{}</file>\n'.format(defect.get('file')))

                # Check for a line number
                if int(defect.get('line')) > 0:
                    output_fh.write('        <line-number>{}</line-number>\n'.format(defect.get('line')))

                output_fh.write('        <comment>{}</comment>\n'.format(defect.get('description')))

                if finding_level.lower() == 'defect':
                    output_fh.write('    </admin_review_defect_create>')
                else:
                    output_fh.write('    </admin_review_comment_create>\n')

        # Add a general comment if there are no uploaded warnings
        if not uploaded_warnings:
            output_fh.write('    <admin_review_comment_create>\n')
            output_fh.write('        <review>{}</review>\n'.format(review_id))
            output_fh.write('        <comment>No warnings were found to include in this review.</comment>\n')
            output_fh.write('    </admin_review_comment_create>\n')
                    
        # Close out the file
        output_fh.write('</batch-commands>\n')


def initialize_upload(tool_conf_data):
    """This function prepares to upload review data to Collaborator.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Create the filtering files
    create_filtering_files(tool_conf_data)

    # Check to see if a user is logged in
    if not check_login(tool_conf_data.get('collaborator_ccollab_location')):
        # Prompt user to log in
        login_flags = tool_conf_data.get('collaborator_server') + ' ' + tool_conf_data.get('collaborator_username')
        login(tool_conf_data.get('collaborator_ccollab_location'), login_flags)

    # Create the review
    subcommand = ('admin review create --creator ' + tool_conf_data.get('collaborator_username') +
                  ' --restrict-access \"' + tool_conf_data.get('collaborator_review_access') + '\" --template \"' +
                  tool_conf_data.get('collaborator_review_template') + '\" --title \"' +
                  tool_conf_data.get('collaborator_review_title') + '\"')
    if tool_conf_data.get('collaborator_review_group'):
        subcommand = (subcommand + ' --group \"' + tool_conf_data.get('collaborator_review_group') + '\"')
    execute_ccollab(tool_conf_data.get('collaborator_ccollab_location'), subcommand)

    # Get the review ID and add it to the configuration data
    review_id = get_review_id(tool_conf_data.get('collaborator_log_file'))
    tool_conf_data.update({'collaborator_review_id': int(review_id)})

    # Set the review xml file path
    files_xml = tool_conf_data.get('scrub_analysis_dir') + '/collaborator_review_' + str(review_id) + '_files.xml'
    comments_xml = tool_conf_data.get('scrub_analysis_dir') + '/collaborator_review_' + str(review_id) + '_comments.xml'
    tool_conf_data.update({'collaborator_review_xml_files': files_xml})
    tool_conf_data.update({'collaborator_review_xml_comments': comments_xml})

    # Set the author of the review
    subcommand = ('admin review set-participants ' + review_id + ' --participant author=' +
                  tool_conf_data.get('collaborator_username'))
    execute_ccollab(tool_conf_data.get('collaborator_ccollab_location'), subcommand)


def perform_upload(tool_conf_data):
    """This function uploads review data to the Collaborator review.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Initialize variables
    tool_defect_types = []
    defect_list = []

    # Get a list of the source code files to be uploaded
    with open(tool_conf_data.get('collaborator_filtering_output_file'), 'r') as fh:
        file_list = list(filter(None, re.split('\n', fh.read())))

    # Check if any files specified
    if tool_conf_data.get('collaborator_src_files'):
        for results_file in tool_conf_data.get('collaborator_src_files').split(','):
            tool_name = results_file[0:-6]
            tool_defect_types.append(tool_name)
    else:
        for root, dirs, files in os.walk(tool_conf_data.get('scrub_analysis_dir')):
            for filename in files:
                if filename.endswith('.scrub') and 'raw' not in filename:
                    tool_name = filename[0:-6]
                    tool_defect_types.append(tool_name)

    # Get the defects
    for tool_name in tool_defect_types:
        scrub_file = os.path.normpath(tool_conf_data.get('scrub_analysis_dir') + '/' + tool_name + '.scrub')

        # Add the defects to the review
        defect_list = defect_list + get_defects(scrub_file)

    # Create XML batch file for uploading files
    create_batch_xml_file_upload(tool_conf_data.get('collaborator_review_xml_files'), file_list,
                                 tool_conf_data.get('collaborator_review_id'),
                                 tool_conf_data.get('collaborator_server'), tool_conf_data.get('collaborator_username'),
                                 tool_conf_data.get('source_dir'))

    # Create the XML batch file for uploading defects
    create_batch_xml_defect_upload(tool_conf_data.get('collaborator_review_xml_comments'), file_list, defect_list,
                                   tool_conf_data.get('collaborator_review_id'),
                                   tool_conf_data.get('collaborator_finding_level'),
                                   tool_conf_data.get('collaborator_server'),
                                   tool_conf_data.get('collaborator_username'))

    # Find the xml files of interest
    xml_files = [tool_conf_data.get('collaborator_review_xml_files'),
                 tool_conf_data.get('collaborator_review_xml_comments')]

    # Perform all the batch commands
    for xml_file in xml_files:
        subcommand = ('admin batch ' + xml_file)
        execute_ccollab(tool_conf_data.get('collaborator_ccollab_location'), subcommand)


def create_filtering_files(tool_conf_data):
    """This function creates the filtering files needed for Collaborator upload.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Print a status message
    logging.info('\t>> Checking analysis filtering output file.')

    # Determine if the baseline filtering list exists, or should be created
    if not os.path.exists(tool_conf_data.get('filtering_output_file')):
        # Print a status message
        logging.info('\t>> No analysis filtering output was found. Creating output now.')

        # Create the output file
        create_file_list.create_file_list(tool_conf_data.get('source_dir'),
                                          tool_conf_data.get('filtering_output_file'),
                                          tool_conf_data.get('analysis_filters'))

    # Create the file list based on filtering options
    create_file_list.create_file_list(tool_conf_data.get('source_dir'),
                                      tool_conf_data.get('collaborator_filtering_output_file'),
                                      tool_conf_data.get('collaborator_filters'),
                                      tool_conf_data.get('filtering_output_file'))


def run_analysis(baseline_conf_data, override=False):
    """This function runs this module based on input from the scrub.cfg file

    Inputs:
        - baseline_conf_data: Dictionary of raw scrub.cfg configuration parameters [dict]
        - override: Force tool execution? [optional] [bool]

    Outputs:
        - collaborator.log: Log file for the Collaborator upload, where uid is the user ID
        - review.xml: XML file containing appropriate Collaborator commands to upload source code and SCRUB results
        - stdout and stderr data will be directed to log_file
    """

    # Import the config file data
    tool_conf_data = baseline_conf_data.copy()
    initialize_analysis(tool_conf_data)

    # Initialize variables
    collaborator_exit_code = 2
    attempt_analysis = tool_conf_data.get('collaborator_upload') or override

    # Determine if Collaborator upload can be run
    if attempt_analysis:
        try:
            # Create the logging file, if it doesn't already exist
            scrub_utilities.create_logger(tool_conf_data.get('collaborator_log_file'))

            # Print a status message
            logging.info('')
            logging.info('Perform Collaborator upload...')

            # Prepare for the upload
            initialize_upload(tool_conf_data)

            # Perform the upload
            perform_upload(tool_conf_data)

            # Set the exit code
            collaborator_exit_code = 0

        except scrub_utilities.CommandExecutionError:
            # Print a warning message
            logging.warning('Collaborator upload could not be performed. Please see log file %s for more information.',
                            tool_conf_data.get('collaborator_log_file'))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            collaborator_exit_code = 1

        except:     # lgtm [py/catch-base-exception]
            # Print a warning message
            logging.warning('Collaborator upload could not be performed. Please see log file %s for more information.',
                            tool_conf_data.get('collaborator_log_file'))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            collaborator_exit_code = 100

        finally:
            # Delete the review if necessary
            if (collaborator_exit_code > 0) and (tool_conf_data.get('collaborator_review_id') > 0):
                subcommand = ('admin review delete ' + str(tool_conf_data.get('collaborator_review_id')))
                execute_ccollab(tool_conf_data.get('collaborator_ccollab_location'), subcommand)

            # Close the loggers
            logging.getLogger().handlers = []

            # Update the permissions of the log file if it exists
            if os.path.exists(tool_conf_data.get('collaborator_log_file')):
                os.chmod(tool_conf_data.get('collaborator_log_file'), 438)

                # Move the log file to line up with the review id, if it exists
                if tool_conf_data.get('collaborator_review_id') > 0:
                    shutil.move(tool_conf_data.get('collaborator_log_file'),
                                tool_conf_data.get('scrub_log_dir') + '/collaborator_' +
                                str(tool_conf_data.get('collaborator_review_id')) + '.log')

    # Return the exit code
    return collaborator_exit_code


def initialize_analysis(tool_conf_data):
    """The purpose of this function is to prepare the tool to perform analysis.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Initialize the derived variables
    current_user = pwd.getpwuid(os.getuid()).pw_name
    collaborator_log_file = os.path.normpath(tool_conf_data.get('scrub_log_dir') + '/collaborator_' +
                                             current_user + '.log')
    collaborator_filtering_output_file = os.path.normpath(tool_conf_data.get('scrub_analysis_dir') +
                                                          '/SCRUBCollaboratorFilteringList')

    # Add derived values to the dictionary
    tool_conf_data.update({'collaborator_log_file': collaborator_log_file})
    tool_conf_data.update({'collaborator_filtering_output_file': collaborator_filtering_output_file})
    tool_conf_data.update({'collaborator_review_id': 0})

    # Set the username if necessary
    if tool_conf_data.get('collaborator_username') == '':
        tool_conf_data.update({'collaborator_username': current_user})

    # Determine if Collaborator can be run
    if not (tool_conf_data.get('collaborator_server')):
        # Update the analysis flag if necessary
        if tool_conf_data.get('collaborator_upload'):
            tool_conf_data.update({'collaborator_upload': False})

            # Print a status message
            print('\nWARNING: Unable to perform Collaborator upload. Required configuration inputs are missing.\n')



import os
import sys
import glob
import argparse
from scrub.tools.parsers import translate_results


def parse_arguments():
    """This function handles argument parsing in preparation for diff utility."""

    # Create the parser
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=diff.__doc__)

    # Add parser arguments
    parser.add_argument('--baseline-source', required=True)
    parser.add_argument('--baseline-scrub', required=True)
    parser.add_argument('--comparison-source', required=True)
    parser.add_argument('--comparison-scrub', required=True)

    # Parse the arguments
    args = vars(parser.parse_args(sys.argv[2:]))

    # Run analysis
    diff(args['baseline_source'], args['baseline_scrub'], args['comparison_source'], args['comparison_scrub'])


def get_lines(source_file, line):
    """This function gets the line number from the source code file of interest

    Inputs:
        - file: Absolute path to the source code file of interest [string]
        - line: Line number of the file to retrieve [int]

    Outputs:
        - source_line: The line of interest from the source code file [string]
    """

    # Initialize variables
    source_lines = []

    # Import the source code data
    with open(source_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Get a 3 lines of source code
    if (line >= 2) and (line <= len(input_data) - 1):
        comparison_source_target = [line - 1, line, line + 1]

    elif line == 1:
        comparison_source_target = [line, line + 1, line + 2]

    elif line >= len(input_data) - 1:
        comparison_source_target = [line - 2, line - 1, line]

    else:
        comparison_source_target = [0]

    # Get the source file line
    for i in comparison_source_target:
        if ((i - 1) <= len(input_data)) and (i != 0):
            source_lines.append(input_data[i - 1].strip())

    return source_lines


def make_warning_relative(warning, source_root):
    """This function makes a warning relative to a root location.

    Inputs:
        - warning: Dictionary containing the warning data [dict]
        - source_root: Absolute path to the source root directory [string]

    Outputs:
        - relative_warning: Dictionary containing the warning data relative to the source root directory [dict]
    """

    # Initialize variables
    relative_warning = warning.copy()

    # Update the file path
    relative_warning['file'] = os.path.relpath(relative_warning['file'], source_root)

    # Update the description
    for line in relative_warning['description']:
        relative_warning['description'][relative_warning['description'].index(line)] = line.replace(source_root + '/',
                                                                                                    '')

    return relative_warning


def find_exact_match(comparison_warning, baseline_warnings):
    """This functions checks to see if an exact match exists.

    Inputs:
        - comparison_warning: Dictionary containing warning data of interest [dict]
        - baseline_warnings: List of warning dictionaries to search [list of dicts]

    Outputs:
        exact_match: Is this an exact match with a baseline warning? [bool]
    """

    # Initialize variables
    exact_match = False

    # See if an exact match can be found
    for baseline_warning in baseline_warnings:
        if ((comparison_warning['file'] == baseline_warning['file']) and
                (comparison_warning['line'] == baseline_warning['line']) and
                (comparison_warning['description'] == baseline_warning['description'])):
            # Print a status message
            print('        >> Exact match found in baseline results: {}'.format(baseline_warning['id']))

            # Update the flag
            exact_match = True

    return exact_match


def find_probable_match(comparison_warning, baseline_warnings, comparison_source_root, baseline_source_root):
    """This function checks to see if a corresponding baseline warning can be found.

    Inputs:
        - comparison_warning: Dictionary containing warning data of interest [dict]
        - baseline_warnings: List of warning dictionaries to search [list of dicts]
        - comparison_source_root: Absolute path to comparison source root directory [string]
        - baseline_source_root: Absolute path to baseline source root directory [string]

    Outputs:
        - probable_match: Is this a probable match with a baseline warning? [bool]
    """

    # Initialize variables
    probable_match = False

    comparison_source_lines = get_lines(comparison_source_root + '/' + comparison_warning['file'],
                                        comparison_warning['line'])

    # Search through all of the baseline warnings and get the corresponding data
    for baseline_warning in baseline_warnings:
        # Check to see if the warning query and warning file match
        if ((baseline_warning['file'] == comparison_warning['file']) and
                (baseline_warning['query'] == comparison_warning['query'])):
            # Get the baseline source file line
            baseline_source_lines = get_lines(baseline_source_root + '/' + baseline_warning['file'],
                                              baseline_warning['line'])

            # Check to see if the source file lines match
            if baseline_source_lines == comparison_source_lines:
                # Update the flag
                probable_match = True

                # Print a status message
                print('        >> Probable match found in baseline results: {}'.format(baseline_warning['id']))

    return probable_match


def diff(baseline_source_root, baseline_scrub_root, comparison_source_root, comparison_scrub_root):
    """
    This function compares a set of static analysis results to a defined baseline set of results.

    Inputs:
        --baseline_source: Absolute path to baseline source root directory [string]
        --baseline_scrub: Absolute path to the baseline SCRUB working directory [string]
        --comparison_source: Absolute path to comparison source root directory [string]
        --comparison_scrub: Absolute path to the comparison SCRUB working directory [string]
    """

    # Find all of the SCRUB files in comparison results
    comparison_scrub_files = glob.glob(comparison_scrub_root + '/*[!_diff].scrub')

    # Iterate through every SCRUB file and remove baseline results
    for comparison_scrub_file in comparison_scrub_files:
        # Print a status message
        print('Examining comparison results file: {}'.format(comparison_scrub_file))

        # Import the comparison SCRUB file data
        comparison_warnings = translate_results.parse_scrub(comparison_scrub_file, comparison_source_root)

        # Make a copy of the results
        comparison_warnings_diff = comparison_warnings.copy()

        # Find the corresponding baseline results file
        baseline_scrub_file = os.path.normpath(baseline_scrub_root + '/' + os.path.basename(comparison_scrub_file))

        # Check to make sure the baseline file exists
        if os.path.exists(baseline_scrub_file):
            # Import the baseline results
            baseline_warnings = translate_results.parse_scrub(baseline_scrub_file, baseline_source_root)

            # Make the baseline warning_relative
            relative_baseline_warnings = []
            for baseline_warning in baseline_warnings:
                relative_baseline_warnings.append(make_warning_relative(baseline_warning, baseline_source_root))

            # Iterate through every warning
            for comparison_warning in comparison_warnings:
                # Make the warning relative
                relative_comparison_warning = make_warning_relative(comparison_warning, comparison_source_root)

                # Print a status message
                print('    >> Searching for warning: {}'.format(comparison_warning['id']))

                # See if an exact match can be found
                exact_match = find_exact_match(relative_comparison_warning, relative_baseline_warnings)

                # See if a probable match can be found
                if exact_match:
                    comparison_warnings_diff.remove(comparison_warning)

                if not exact_match:
                    # Print a status message
                    print('        >> No exact match found. Attempting to find probable match.')

                    probable_match = find_probable_match(relative_comparison_warning, relative_baseline_warnings,
                                                         comparison_source_root, baseline_source_root)

                    # Print a status message
                    if probable_match:
                        comparison_warnings_diff.remove(comparison_warning)

                    if not probable_match:
                        print('        >> No probable match found.')

        else:
            # Print a warning message
            print('    >> Baseline results file {} does not exist.'.format(baseline_scrub_file))
            print('    >> All results are new.')

        # Create the output file path
        diff_output_file = (comparison_scrub_root + '/' +
                            os.path.basename(comparison_scrub_file).replace('.scrub', '') + '_diff.scrub')

        # Write out the results
        translate_results.create_scrub_output_file(comparison_warnings_diff, diff_output_file)



import glob
import importlib
import os
import re
import shutil
import sys
import time
import argparse
import logging
import traceback
from scrub.utils.filtering import do_filtering
from scrub.utils import do_clean
from scrub.utils import scrub_utilities


def parse_arguments():
    """This function parse command line arguments in preparation for analysis."""

    # Create the parser
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=main.__doc__)

    # Add parser arguments
    parser.add_argument('--config', default='./scrub.cfg')
    parser.add_argument('--clean', action='store_true')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--quiet', action='store_true')
    parser.add_argument('--tools', nargs='+', default=None)
    parser.add_argument('--targets', nargs='+', default=None)

    # Parse the arguments
    args = vars(parser.parse_args(sys.argv[2:]))

    # Set the logging level
    if args['debug']:
        logging_level = logging.DEBUG
    elif args['quiet']:
        logging_level = logging.CRITICAL
    else:
        logging_level = logging.INFO

    # Run analysis
    main(args['config'], args['clean'], logging_level, args['tools'], args['targets'])


def main(conf_file='./scrub.cfg', clean=False, console_logging=logging.INFO, tools=None, targets=None):
    """
    This function runs all applicable tools present within the configuration file.

    Inputs:
        - config: Path to SCRUB configuration file [string] [optional]
            Default value: ./scrub.cfg
        - clean: Should SCRUB clean existing results? [bool]
            Default value: False
        - console_logging: Logging level for console [int] [optional]
            Default value: logging.INFO (20)
        - tools: List of tools to run during analysis [list of strings] [optional]
            Default value: None
        - targets: List of output targets for exporting the analysis results [list of strings] [optional]
            Default value: None
    """

    # Read in the configuration data
    if os.path.exists(conf_file):
        scrub_conf_data = scrub_utilities.parse_common_configs(conf_file)
    else:
        print('ERROR: Configuration file ' + conf_file + ' does not exist.')
        sys.exit(10)

    # Initialize variables
    execution_status = []
    scrub_path = os.path.dirname(os.path.realpath(__file__))

    # Clean the previous SCRUB data from the current directory
    if clean:
        do_clean.clean_directory(scrub_conf_data.get('source_dir'))

    # Initialize the SCRUB storage directory
    scrub_utilities.initialize_storage_dir(scrub_conf_data)

    # Make a copy of the scrub.cfg file and add it to the log
    shutil.copyfile(conf_file, os.path.normpath(scrub_conf_data.get('scrub_analysis_dir') + '/scrub.cfg'))

    try:
        # Handle legacy language selections
        if scrub_conf_data.get('source_lang') == 'j':
            scrub_conf_data.update({'source_lang': 'java'})
        if scrub_conf_data.get('source_lang') == 'p':
            scrub_conf_data.update({'source_lang': 'python'})

        # Get the templates for the language
        available_analysis_templates = glob.glob(scrub_path + '/tools/templates/' + scrub_conf_data.get('source_lang') +
                                                 '/*.template')

        # Append the custom templates if provided
        if scrub_conf_data.get('custom_templates'):
            available_analysis_templates = (available_analysis_templates +
                                            scrub_conf_data.get('custom_templates').replace('\"', '').split(','))

        # Check to make sure at least one possible template has been identified
        if len(available_analysis_templates) == 0:
            print('WARNING: No analysis templates have been found.')

        # Update the analysis templates to be run
        if tools == 'filter' or tools == 'filtering':
            analysis_templates = []
        elif tools:
            analysis_templates = []
            for template in available_analysis_templates:
                for tool in tools:
                    if template.endswith(tool + '.template'):
                        analysis_templates.append(template)
                        scrub_conf_data.update({tool.lower() + '_warnings': True})
        else:
            analysis_templates = available_analysis_templates

        # Perform analysis using the template
        for analysis_template in analysis_templates:
            # Initialize variables
            execution_time = 0

            # Get the tool name
            tool_name = os.path.splitext(os.path.basename(analysis_template))[0]

            # Initialize execution status
            tool_execution_status = 2

            if scrub_conf_data.get(tool_name.lower() + '_warnings'):
                # Initialize variables
                analysis_scripts_dir = os.path.normpath(scrub_conf_data.get('scrub_analysis_dir') + '/analysis_scripts')
                analysis_script = os.path.normpath(analysis_scripts_dir + '/' + tool_name + '.sh')
                tool_analysis_dir = os.path.normpath(scrub_conf_data.get('scrub_working_dir') + '/' + tool_name +
                                                     '_analysis')

                # Add derived values to configuration values
                scrub_conf_data.update({'tool_analysis_dir': tool_analysis_dir})

                # Create the analysis scripts directory
                if not os.path.exists(analysis_scripts_dir):
                    os.mkdir(analysis_scripts_dir)

                # Create the tool analysis directory
                if os.path.exists(tool_analysis_dir):
                    shutil.rmtree(tool_analysis_dir)
                os.mkdir(tool_analysis_dir)

                # Create the log file
                analysis_log_file = scrub_conf_data.get('scrub_log_dir') + '/' + tool_name + '.log'
                scrub_utilities.create_logger(analysis_log_file, console_logging)

                # Print a status message
                logging.info('')
                logging.info('  Configuration values...')

                # Print general configuration values
                logging.info('    SOURCE_DIR: ' + scrub_conf_data.get('source_dir'))
                logging.info('    TOOL_ANALYSIS_DIR: ' + scrub_conf_data.get('scrub_working_dir') + '/' + tool_name +
                             '_analysis')

                # Print tool specific configuration values
                for config_value in scrub_conf_data.keys():
                    if config_value.startswith(tool_name):
                        logging.info('    ' + config_value.upper() + ': ' + str(scrub_conf_data.get(config_value)))

                # Print a status message
                logging.info('')
                logging.info('  Parsing ' + tool_name + ' template file...')

                # Create the analysis template
                scrub_utilities.parse_template(analysis_template, analysis_script, scrub_conf_data)

                # Update the permissions of the analysis script
                # os.chmod(analysis_script, 0o777)

                try:
                    # Start the timer
                    start_time = time.time()

                    # Set the environment for execution
                    user_env = os.environ.copy()

                    # Update the environment
                    if 'PYTHONPATH' in user_env.keys():
                        user_env.update({'PYTHONPATH': user_env.get('PYTHONPATH') + ':' +
                                        os.path.normpath(scrub_path + '/../')})
                    else:
                        user_env.update({'PYTHONPATH': os.path.normpath(scrub_path + '/../')})

                    # Execute the analysis and track execution time
                    scrub_utilities.execute_command(analysis_script, user_env)

                    # Check the tool analysis directory
                    scrub_utilities.check_artifact(tool_analysis_dir, True)

                    # Check the raw results files
                    for raw_results_file in glob.glob(os.path.join(scrub_conf_data['raw_results_dir'],
                                                                   tool_name + '_*.scrub')):
                        scrub_utilities.check_artifact(raw_results_file, False)

                    # Check the log file
                    scrub_utilities.check_artifact(analysis_log_file, True)

                    # Update the execution status
                    tool_execution_status = 0

                except scrub_utilities.CommandExecutionError:
                    logging.warning(tool_name + ' analysis could not be performed.')

                    # Print the exception traceback
                    logging.warning(traceback.format_exc())

                    #  Update the execution status
                    tool_execution_status = 1

                finally:
                    # Close the logger
                    logging.getLogger().handlers = []

                    # Calculate the execution time
                    execution_time = time.time() - start_time

            # Update the execution status
            execution_status.append([tool_name, tool_execution_status, execution_time])

        # Perform filtering and track execution time
        start_time = time.time()
        filtering_status = do_filtering.run_analysis(scrub_conf_data)
        execution_time = time.time() - start_time

        # Update the execution status
        execution_status.append(['filtering', filtering_status, execution_time])

    finally:
        # Move the results back with the source code if necessary
        if scrub_conf_data.get('scrub_working_dir') != scrub_conf_data.get('scrub_analysis_dir'):
            # Move every item in the directory
            for item in os.listdir(scrub_conf_data.get('scrub_working_dir')):
                # Remove the destination directory, if it exists
                if os.path.exists(scrub_conf_data.get('scrub_analysis_dir') + '/' + item):
                    shutil.rmtree(scrub_conf_data.get('scrub_analysis_dir') + '/' + item)

                # Move the contents
                shutil.move(scrub_conf_data.get('scrub_working_dir') + '/' + item,
                            scrub_conf_data.get('scrub_analysis_dir') + '/' + item)

            # Remove the working directory
            shutil.rmtree(scrub_conf_data.get('scrub_working_dir'))

        # Print a status message
        tool_failure_count = 0
        total_execution_time = 0
        print('\nTool Execution Status:\n')
        for status in execution_status:
            # Track the execution time
            total_execution_time = total_execution_time + status[2]

            # Decode the exit code
            if status[1] == 0:
                exit_code = 'Completed successfully'
            elif status[1] == 1:
                exit_code = 'Attempted analysis, did not complete'
                tool_failure_count = tool_failure_count + 1
            elif status[1] == 2:
                exit_code = 'Not attempted'
            elif status[1] == 100:
                exit_code = 'Fatal error'
            else:
                exit_code = 'Unknown error'

            # Print the status message
            print('\t%s | %s: %s' % (time.strftime("%H:%M:%S", time.gmtime(status[2])), status[0], exit_code))
            print('\t--------------------------------------------------')

        # Print the execution time
        print('\n\tTotal Execution Time: %s\n' % time.strftime("%H:%M:%S", time.gmtime(total_execution_time)))

    # Search for target modules
    available_target_modules = glob.glob(scrub_path + '/targets/*/do_*.py')

    # Handle legacy Collaborator tag
    if 'collaborator_upload' in scrub_conf_data.keys():
        scrub_conf_data.update({'collaborator_export': scrub_conf_data.get('collaborator_upload')})
    else:
        scrub_conf_data.update({'collaborator_export': False})

    # Update the targets to be run
    if targets:
        target_modules = []
        for module_path in available_target_modules:
            for target in targets:
                if target in os.path.basename(module_path):
                    target_modules.append(module_path)
                    scrub_conf_data.update({target.lower() + '_export': True})
    else:
        target_modules = available_target_modules

    # Loop through every tool and perform
    for target_module in target_modules:
        # Form the call string
        module_name = 'scrub.' + re.split('\\.py', os.path.relpath(target_module, scrub_path))[0].replace('/', '.')

        # Import the module
        module_object = importlib.import_module(module_name)

        # Call the analysis
        getattr(module_object, "run_analysis")(scrub_conf_data)

    # Set the exit code
    sys.exit(tool_failure_count)



import os
import re
import sys
import json
import logging
from distutils.version import StrictVersion
from scrub.tools.parsers import translate_results

WARNING_LEVEL = 'Low'
ID_PREFIX = 'klocwork'


def parse_warnings_2020_2(input_file, output_file):
    """This function parse Klocwork results from v2020.2.

    Inputs:
        - input_file: Absolute path to the file containing raw Klocwork warnings [string]
        - output_file: Absolute path to the file where the parsed warnings will be stored [string]
    """

    # Print status message
    logging.info('')
    logging.info('\t>> Executing command: get_klocwork_warnings.parse_warnings_2020_2(%s, %s)', input_file,
                 output_file)
    logging.info('\t>> From directory: %s', os.getcwd())

    # Initialize the variables
    warning_count = 1

    # Import the input data file
    with open(input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Create the output file
    raw_warnings = []
    for line in input_data:
        # Convert the warning to a dictionary
        warning_data = json.loads(line)

        # Parse the warning data
        warning_file = warning_data['file']
        warning_line = 0
        warning_query = warning_data['code']
        warning_message = warning_data['message']
        warning_link = re.split(',searchquery', warning_data['url'])[0]
        warning_id = ID_PREFIX + str(warning_count).zfill(3)
        warning_description = [warning_message, warning_link]

        # Add to the warning dictionary
        raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                             warning_description, ID_PREFIX, WARNING_LEVEL,
                                                             warning_query))

        # Increment the warning count
        warning_count = warning_count + 1

    # Create the SCRUB output file
    translate_results.create_scrub_output_file(raw_warnings, output_file)


def parse_warnings(input_file, output_file, version_number):
    """This function will examine the input_file to determine which parser will be used.

    Inputs:
        - input_file: Absolute path to the file containing raw Klocwork warnings [string]
        - output_file: Absolute path to the file where the parsed warnings will be stored [string]
        - version_number: Version number for Klocwork instance being used [string]

    Outputs:
        - output_file: All parsed warnings will be written to the output_file
    """

    if StrictVersion(version_number) >= StrictVersion('20.2.0'):
        parse_warnings_2020_2(input_file, output_file)
    else:
        parse_warnings_2020_2(input_file, output_file)


if __name__ == '__main__':
    parse_warnings(sys.argv[1], sys.argv[2], sys.argv[3])



import re
import os
import sys
import logging
from distutils.version import StrictVersion

WARNING_LEVEL = 'Low'
ID_PREFIX = 'coverity'


def get_error_indices(raw_input_file):
    """This function gets the indices of the first line of all Coverity warnings.

    Inputs:
        - raw_input_file: Full path to the file containing raw Coverity warnings [string]

    Outputs:
        - error_indices: List of warning indices [list of int]
    """

    # Initialize variables
    error_indices = []

    # Import the input data file
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Iterate through every line of the input file
    for i in range(0, len(input_data)):
        # Get the line
        line = input_data[i].strip()

        if ('Error:' in line) or ('Checker:' in line) or ('Type:' in line):
            error_indices.append(i)

    return error_indices


def parse_warnings_2019_12(raw_input_file, parsed_output_file):
    """This function parses the raw Coverity warnings (version 2019.12) into the SCRUB format.

    Inputs:
        - input_file: Absolute path to the file containing raw Coverity warnings [string]
        - parsed_output_file: Absolute path to the file where the parsed warnings will be stored [string]
    """

    # Print status message
    logging.info('')
    logging.info('\t>> Executing command: get_coverity_warnings.parse_warnings_2019_12(%s, %s)',
                 raw_input_file, parsed_output_file)
    logging.info('\t>> From directory: %s', os.getcwd())

    # Import the input data file
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Create the output file
    with open(parsed_output_file, 'w+') as output_fh:
        # Iterate through every line of the input file
        error_indices = get_error_indices(raw_input_file)

        # Iterate through every line of the input file and parse warnings
        for i in range(0, len(error_indices)):
            # Initialize variables
            warning_text = []

            # Get the index line
            error_index = error_indices[i]

            # Get the name of the warnings
            warning_name = list(filter(None, re.split('[()]', input_data[error_index].strip())))[-1].strip()

            # Get the location information
            line = input_data[error_index - 1].strip()
            line_split = list(filter(None, re.split(':', line)))
            warning_file = line_split[-2]
            warning_line = int(line_split[-1])

            # Increment the warning count
            warning_count = i + 1

            # Get the warning text
            if i < len(error_indices) - 1:
                warning_index_end = error_indices[i + 1] - 2
            else:
                warning_index_end = len(input_data)

            for j in range(error_index + 1, warning_index_end):
                # Add the line ot the list, if it's not blank
                if not input_data[j].strip() == '':
                    warning_text.append(input_data[j].strip())

            # Write the data to the output file
            output_fh.write('%s%03d <%s> :%s:%d: %s\n' % (ID_PREFIX, warning_count, WARNING_LEVEL, warning_file,
                                                          warning_line, warning_name))
            for line in warning_text:
                output_fh.write('    %s\n' % line)
            output_fh.write('\n')

    # Change the permissions of the output file
    os.chmod(parsed_output_file, 438)


def parse_warnings_2019_06(raw_input_file, parsed_output_file):
    """This function parses the raw Coverity warnings (version 2019.06) into the SCRUB format.

    Inputs:
        - raw_input_file: Full path to the file containing raw Coverity warnings [string]
        - parsed_output_file: Full path to the file where the parsed warnings will be stored [string]
    """

    # Print status message
    logging.info('')
    logging.info('\tParsing results...')
    logging.info('\t>> Executing command: get_coverity_warnings.parse_warnings_2019_06(%s, %s)',
                 raw_input_file, parsed_output_file)
    logging.info('\t>> From directory: %s', os.getcwd())

    # Import the input data file
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Create the output file
    with open(parsed_output_file, 'w+') as output_fh:
        # Iterate through every line of the input file
        error_indices = get_error_indices(raw_input_file)

        # Iterate through every line of the input file and parse warnings
        for i in range(0, len(error_indices)):
            # Initialize variables
            warning_text = []

            # Get the index line
            error_index = error_indices[i]

            # Get the name of the warnings
            warning_name = list(filter(None, re.split(':', input_data[error_index].strip())))[-1].strip()

            # Get the location information
            line = input_data[error_index - 1].strip()
            line_split = list(filter(None, re.split(':', line)))
            warning_file = line_split[-2]
            warning_line = int(line_split[-1])

            # Increment the warning count
            warning_count = i + 1

            # Get the warning text
            if i < len(error_indices) - 1:
                warning_index_end = error_indices[i + 1] - 2
            else:
                warning_index_end = len(input_data)

            for j in range(error_index + 1, warning_index_end):
                # Add the line ot the list, if it's not blank
                if not input_data[j].strip() == '':
                    warning_text.append(input_data[j].strip())

            # Write the data to the output file
            output_fh.write('%s%03d <%s> :%s:%d: %s\n' % (ID_PREFIX, warning_count, WARNING_LEVEL, warning_file,
                                                          warning_line, warning_name))
            for line in warning_text:
                output_fh.write('    %s\n' % line)
            output_fh.write('\n')

    # Change the permissions of the output file
    os.chmod(parsed_output_file, 438)


def parse_warnings_legacy(raw_input_file, parsed_output_file):
    """This function parses the raw Coverity warnings (version 2018.09 and older) into the SCRUB format.

    Inputs:
        - raw_input_file: Full path to the file containing raw Coverity warnings [string]
        - parsed_output_file: Full path to the file where the parsed warnings will be stored [string]
    """

    # Print status message
    logging.info('')
    logging.info('\tParsing results...')
    logging.info('\t>> Executing command: get_coverity_warnings.parse_warnings_legacy(%s, %s)',
                 raw_input_file, parsed_output_file)
    logging.info('\t>> From directory: %s', os.getcwd())

    # Import the input data file
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Create the output file
    with open(parsed_output_file, 'w+') as output_fh:
        # Iterate through every line of the input file
        error_indices = get_error_indices(raw_input_file)

        # Iterate through every line of the input file and parse warnings
        for i in range(0, len(error_indices)):
            # Initialize variables
            warning_text = []

            # Get the index line
            error_index = error_indices[i]

            # Get the name of the warnings
            warning_name = list(filter(None, re.split(':', input_data[error_index].strip())))[-1].strip()

            # Get the location information
            line = input_data[error_index + 1].strip()
            line_split = list(filter(None, re.split(':', line)))
            warning_file = line_split[-2]
            warning_line = int(line_split[-1])

            # Increment the warning count
            warning_count = i + 1

            # Get the warning text
            if i < len(error_indices)-1:
                warning_index_end = error_indices[i+1]-1
            else:
                warning_index_end = len(input_data)

            for j in range(error_index+1, warning_index_end):
                # Add the line ot the list
                warning_text.append(input_data[j].strip())

            # Write the data to the output file
            output_fh.write('%s%03d <%s> :%s:%d: %s\n' % (ID_PREFIX, warning_count, WARNING_LEVEL, warning_file,
                                                          warning_line, warning_name))
            for line in warning_text:
                output_fh.write('    %s\n' % line)
            output_fh.write('\n')

    # Change the permissions of the output file
    os.chmod(parsed_output_file, 438)


def parse_warnings(raw_input_file, parsed_output_file, coverity_version_number):
    """This function will examine the raw_input_file to determine which parser will be used.

    Inputs:
        - raw_input_file: Full path to the file containing raw Coverity warnings [string]
        - parsed_output_file: Full path to the file where the parsed warnings will be stored [string]
        - version_number: Version number for Coverity instance being used [string]
    """

    # Select which parser should be used
    if StrictVersion(coverity_version_number) >= StrictVersion('2019.12'):
        parse_warnings_2019_12(raw_input_file, parsed_output_file)
    elif (StrictVersion(coverity_version_number) >= StrictVersion('2019.06')) and \
            (StrictVersion(coverity_version_number) < StrictVersion('2019.12')):
        parse_warnings_2019_06(raw_input_file, parsed_output_file)
    else:
        parse_warnings_legacy(raw_input_file, parsed_output_file)


if __name__ == '__main__':
    parse_warnings(sys.argv[1], sys.argv[2], sys.argv[3])



import sys
import xml.etree.ElementTree
import os
import re
import logging


def find_table_indices(input_file):
    """This function finds the proper table instances for warning attributes.

    Inputs:
        - input_file: Absolute path to the raw warning file for the CodeSonar Hub [string]

    Outputs:
        - indices: List of indices for the desired warning values [list of int]
    """

    # Initialize the variables
    indices = [None, None, None, None]

    # Import the input data
    codesonar_data = xml.etree.ElementTree.parse(input_file).getroot()

    # Iterate through the output file to find the column name data
    for table in codesonar_data.findall('table'):
        # Check the name of the table
        if table.attrib['name'] == 'Warnings':
            index = 0
            for header in table.findall('column'):

                # Find the header indices
                if header.attrib['name'] == 'Path':
                    indices[0] = index
                elif header.attrib['name'] == 'Line':
                    indices[1] = index
                elif header.attrib['name'] == 'Class':
                    indices[2] = index
                elif header.attrib['name'] == 'Procedure':
                    indices[3] = index

                # Increment the index
                index = index + 1

    return indices


def get_hub_location(input_file):
    """This function returns the CodeSonar Hub location from the raw results file.

    Inputs:
        - input_file: Absolute path to the raw warning file for the CodeSonar Hub [string]

    Outputs:
        - codesonar_hub_location: The CodeSonar Hub location found in the raw results file. [string]
    """

    # Initialize variables
    codesonar_hub_location = None

    # Import the input data
    codesonar_data = xml.etree.ElementTree.parse(input_file).getroot()

    # Iterate through the output file to find the column name data
    for text in codesonar_data.findall('text'):
        if text.attrib['class'] == 'Title1':
            codesonar_hub_location = text.attrib['value']
            break

    return codesonar_hub_location


def parse_warnings(input_file, output_file, exclude_p10=False):
    """This function parses the raw CodeSonar warnings into the SCRUB format.

    Inputs:
        - input_file: Full path to the file containing raw CodeSonar warnings [string]
        - output_file: Full path to the file where the parsed warnings will be stored [string]

    Outputs:
        - output_file: All parsed warnings will be written to the output_file
    """

    # Print a status message
    logging.info('\t>> Executing command: get_codesonar_warnings.parse_warnings(%s, %s)', input_file, output_file)
    logging.info('\t>> From directory: %s', os.getcwd())

    # Initialize the variables
    warning_level = 'Low'
    warning_count = 1
    id_prefix = 'codesonar'
    p10_only = False
    filter_list = ["Goto Statement",  # P10: Rule 1
                   "Recursion",  # P10: Rule 1
                   "Use of < setjmp.h >",  # P10: Rule 1
                   "Use of longjmp",  # P10: Rule 1
                   "Use of setjmp",  # P10: Rule 1
                   "Potential Unbounded Loop",  # P10: Rule 2
                   "Dynamic Allocation After Initialization",  # P10: Rule 3
                   "Function Too Long",  # P10: Rule 4
                   "Not Enough Assertions",  # P10: Rule 5
                   "Scope Could Be File Static",  # P10: Rule 6
                   "Scope Could Be Local Static",  # P10: Rule 6
                   "Ignored Return Value",  # P10: Rule 7
                   "Unchecked Parameter Dereference",  # P10: Rule 7
                   "## Follows # Operator",  # P10: Rule 8
                   "Conditional Compilation",  # P10: Rule 8
                   "Macro Does Not End With } or )",  # P10: Rule 8
                   "Macro Does Not Start With { or (",  # P10: Rule 8
                   "Macro Name is C Keyword",  # P10: Rule 8
                   "Macro Uses # Operator",  # P10: Rule 8
                   "Macro Uses ## Operator",  # P10: Rule 8
                   "Non-Boolean Preprocessor Expression",  # P10: Rule 8
                   "Preprocessing Directives in Macro Argument",  # P10: Rule 8
                   "Recursive Macro",  # P10: Rule 8
                   "Unbalanced Parenthesis",  # P10: Rule 8
                   "Use of <stdio.h> Input/Output Macro",  # P10: Rule 8
                   "Use of <wchar.h> Input/Output Macro",  # P10: Rule 8
                   "Variadic Macro",  # P10: Rule 8
                   "Function Pointer",  # P10: Rule 9
                   "Macro Uses [] Operator",  # P10: Rule 9
                   "Macro Uses -> Operator",  # P10: Rule 9
                   "Macro Uses Unary * Operator",  # P10: Rule 9
                   "Pointer Type Inside Typedef",  # P10: Rule 9
                   "Too Many Dereferences",  # P10: Rule 9
                   "Not All Warnings Are Enabled",  # P10: Rule 10
                   "Warnings Not Treated As Errors"]  # P10: Rule 10

    # Determine if the output is P10 results
    if 'p10' in os.path.basename(output_file).lower():
        p10_only = True

    # Get the header indices
    header_indices = find_table_indices(input_file)

    # Get the CodeSonar hub location
    codesonar_hub_location = get_hub_location(input_file)

    # Set the indices for retrieving data
    path_index = header_indices[0]
    line_index = header_indices[1]
    class_index = header_indices[2]
    procedure_index = header_indices[3]

    # Import the input data
    codesonar_data = xml.etree.ElementTree.parse(input_file).getroot()

    # Create the output file
    with open(output_file, 'w+') as output_fh:
        # Find all the tables in the XML file
        for table in codesonar_data.findall('table'):
            # Check the name of the table
            if table.attrib['name'] == 'Warnings':
                # Iterate through all the warnings in the table
                for warning in table.findall('row'):
                    # Parse the warning into SCRUB format
                    warning_instance_id = re.split('\.', warning[0].text)[-1]
                    warning_file = warning[path_index].text
                    warning_line = int(warning[line_index].text)
                    warning_class = warning[class_index].text
                    if str(warning[procedure_index].text).lower() == 'none':
                        warning_procedure = 'undefined procedure'
                    else:
                        warning_procedure = warning[procedure_index].text
                    warning_summary = warning_class + ' found in ' + warning_procedure
                    warning_link = codesonar_hub_location + '/warninginstance/' + warning_instance_id + '.html'

                    # Check to see if the warnings should be filtered for P10 results
                    if p10_only:
                        # Check to see if the warning class is associated with P10 warnings
                        if warning_class.strip() in filter_list:
                            # Write the warnings to the output file
                            output_fh.write('%s%03d <%s> :%s:%d: %s\n\tCodeSonar P10 Warning: %s\n\t%s\n\n' %
                                            (id_prefix, warning_count, warning_level, warning_file, warning_line,
                                             warning_class, warning_summary, warning_link))

                            # Increase the warning_count
                            warning_count = warning_count + 1

                    elif exclude_p10:
                        # Check to see if the warning class is associated with P10 warnings
                        if warning_class.strip() not in filter_list:
                            # Write the warnings to the output file
                            output_fh.write('%s%03d <%s> :%s:%d: %s\n\t%s\n\t%s\n\n' %
                                            (id_prefix, warning_count, warning_level, warning_file, warning_line,
                                             warning_class, warning_summary, warning_link))

                            # Increase the warning_count
                            warning_count = warning_count + 1

                    else:
                        # Write the warnings to the output file
                        output_fh.write('%s%03d <%s> :%s:%d: %s\n\t%s\n\t%s\n\n' %
                                        (id_prefix, warning_count, warning_level, warning_file, warning_line,
                                         warning_class, warning_summary, warning_link))

                        # Increase the warning_count
                        warning_count = warning_count + 1

    # Update the permissions of the output file
    os.chmod(output_file, 438)


if __name__ == '__main__':
    # Store input variables
    codesonar_file = sys.argv[1]
    scrub_file = sys.argv[2]

    parse_warnings(codesonar_file, scrub_file)
